from mopidy import ext, config
import pathlib
from .backend import YandexMusicBackend

__version__ = '0.0.6'


class Extension(ext.Extension):
    dist_name = "Mopidy-YandexMusic"
    ext_name = "yandexmusic"
    version = __version__

    def get_default_config(self):
        default_config = config.read(pathlib.Path(__file__).parent / "ext.conf")
        return default_config

    def get_config_schema(self):
        schema = super().get_config_schema()
        schema["login"] = config.String()
        schema["password"] = config.Secret()
        schema["bitrate"] = config.Integer(optional=True)
        return schema

    def validate_config(self, config):
        return True

    def setup(self, registry):
        registry.add("backend", YandexMusicBackend)



from mopidy import backend, audio
from yandex_music import Client


class YandexMusicPlaybackProvider(backend.PlaybackProvider):
    def __init__(self, client: Client, audio: audio.Audio, backend: backend.Backend, bitrate:int):
        super().__init__(audio, backend)
        self._client = client
        self._bitrate = bitrate

    def translate_uri(self, uri: str):
        _, kind, artist_id, track_id = uri.split(":")
        uid = f"{artist_id}:{track_id}"
        infos = self._client.tracks_download_info(uid, get_direct_links=True)
        for info in infos:
            if info.codec == "mp3" and info.bitrate_in_kbps == self._bitrate:
                link = info.direct_link
                return link
        return None



from mopidy import backend, models
from yandex_music import Client
from .caches import YMTrackCache
from .classes import YMTrack


class YandexMusicLibraryProvider(backend.LibraryProvider):
    def __init__(self, client: Client, track_cache: YMTrackCache):
        self._client = client
        self._track_cache = track_cache
        self.root_directory = models.Ref.directory(uri="yandexmusic:directory:root", name="root")

    def browse(self, uri):
        return None

    def lookup(self, uri: str):
        track = self._track_cache.get(uri)
        if track is not None:
            return [track]

        _, kind, ymartist_id, ymtrack_id = uri.split(":")
        track_id = f"{ymartist_id}:{ymtrack_id}"
        ymtrack = self._client.tracks(track_id)

        track = YMTrack.from_track(ymtrack)
        return [track]

    def get_images(self, uris):
        result = dict()

        for uri in uris:
            _, kind, id = uri.split(":", 2)
            if kind == "track":
                track = self._track_cache.get(uri)
                if track is None:
                    continue
                artwork_uri = "https://" + track.artwork.replace("%%", "400x400")
                result[uri] = [models.Image(uri=artwork_uri)]
            if kind == "playlist":
                pass
        return result



from .classes import YMTrack


class YMTrackCache:
    def __init__(self):
        self._cache = dict()

    def put(self, track: YMTrack):
        self._cache[track.uri] = track

    def get(self, uri: str) -> YMTrack:
        if uri not in self._cache:
            return None

        track = self._cache[uri]
        return track



from mopidy import backend, audio
import pykka
import yandex_music
from .playlist_provider import YandexMusicPlaylistProvider
from .playback_provider import YandexMusicPlaybackProvider
from .library_provider import YandexMusicLibraryProvider
from .caches import YMTrackCache


class YandexMusicBackend(pykka.ThreadingActor, backend.Backend):
    def __init__(self, config: dict, audio: audio):
        super().__init__()

        ym_config :dict = config["yandexmusic"]

        login = ym_config["login"]
        password = ym_config["password"]
        bitrate = int(ym_config["bitrate"]) if "bitrate" in ym_config else 192

        self._config = config
        self._audio = audio

        client = yandex_music.Client.from_credentials(username=login, password=password)
        track_cache = YMTrackCache()

        self.playlists = YandexMusicPlaylistProvider(client, track_cache)
        self.playback = YandexMusicPlaybackProvider(client, audio, self, bitrate)
        self.library = YandexMusicLibraryProvider(client, track_cache)

        self.uri_schemes = ["yandexmusic"]



from mopidy import backend
import yandex_music
from .classes import YMRef, YMPlaylist
from typing import List
from .caches import YMTrackCache

_YM_GENERATED = "yandex-generated"


class YandexMusicPlaylistProvider(backend.PlaylistsProvider):
    def __init__(self, client: yandex_music.Client, track_cache: YMTrackCache):
        self._client = client
        self._track_cache = track_cache

    def as_list(self) -> List[YMRef]:
        yandex_daily = YMRef.from_raw(_YM_GENERATED, "yamusic-daily", "Дневная предложка от Яндекса")
        yandex_alice = YMRef.from_raw(_YM_GENERATED, "yamusic-origin", "Предложка от Алисы")
        yandex_dejavu = YMRef.from_raw(_YM_GENERATED, "yamusic-dejavu", "Дежавю от Яндекса")

        playlists = self._client.users_playlists_list()
        refs = list(map(YMRef.from_playlist, playlists))
        refs.extend([yandex_daily, yandex_alice, yandex_dejavu])
        return refs

    def get_items(self, uri: str) -> YMRef:
        _, kind, ym_userid, playlist_id = uri.split(":")

        if ym_userid == str(self._client.me.account.uid):
            playlist = self._client.users_playlists(playlist_id)[0]
            track_ids = list(map(lambda t: t.track_id, playlist.tracks))
            tracks = self._client.tracks(track_ids)
            refs = list(map(YMRef.from_track, tracks))
            return refs

    def lookup(self, uri: str) -> YMPlaylist:
        _, kind, ym_userid, playlist_id = uri.split(":")
        try:
            if ym_userid == str(self._client.me.account.uid):
                ymplaylist = self._client.users_playlists(playlist_id)[0]
                track_ids = list(map(lambda t: t.track_id, ymplaylist.tracks))
                ymplaylist.tracks = self._client.tracks(track_ids)
                playlist = YMPlaylist.from_playlist(ymplaylist)
                for track in playlist.tracks:
                    self._track_cache.put(track)
                return playlist
            elif ym_userid == _YM_GENERATED:
                '''
                Костыль!!!
                users_playlists_list для генерированного контента возвращает плейлист без треков.
                Для получения полноценного плейлиста вызываем users_playlists c айдишником полученного плейлиста                    
                '''
                foreign_playlist = self._client.users_playlists_list(user_id=playlist_id)
                ymplaylists = self._client.users_playlists(kind=foreign_playlist[0].kind, user_id=playlist_id)
                ymplaylist = ymplaylists[0]
                track_ids = list(map(lambda t: t.track_id, ymplaylist.tracks))
                ymplaylist.tracks = self._client.tracks(track_ids)
                playlist = YMPlaylist.from_playlist(ymplaylist)
                for track in playlist.tracks:
                    self._track_cache.put(track)
                return playlist
            else:
                ymplaylist = self._client.users_playlists(playlist_id, user_id=ym_userid)[0]
                track_ids = list(map(lambda t: t.track_id, ymplaylist.tracks))
                ymplaylist.tracks = self._client.tracks(track_ids)
                playlist = YMPlaylist.from_playlist(ymplaylist)
                for track in playlist.tracks:
                    self._track_cache.put(track)
                return playlist

        except Exception as e:
            print("Exception")

    def create(self, name):
        return None

    def delete(self, uri):
        return None

    def refresh(self):
        pass

    def save(self, playlist):
        return None



from mopidy.models import Playlist, Track, Ref, fields, Artist
from yandex_music import Playlist as YMPlaylist, Track as YMTrack, Artist as YArtist


class YMRef(Ref):
    @staticmethod
    def from_raw(owner: str, playlist_id: str, title: str):
        uri = f"yandexmusic:playlist:{owner}:{playlist_id}"
        name = title
        ref = YMRef(type=Ref.PLAYLIST, uri=uri, name=name)

        return ref

    @staticmethod
    def from_playlist(playlist: YMPlaylist):
        uri = f"yandexmusic:playlist:{playlist.playlist_id}"
        name = playlist.title
        ref = YMRef(type=Ref.PLAYLIST, uri=uri, name=name)

        return ref

    @staticmethod
    def from_track(track: YMTrack):
        uri = f"yandexmusic:track:{track.track_id}"
        name = track.title
        ref = YMRef(type=Ref.TRACK, uri=uri, name=name)

        return ref


class YMPlaylist(Playlist):
    @staticmethod
    def from_playlist(playlist: YMPlaylist):
        uri = f"yandexmusic:playlist:{playlist.playlist_id}"
        name = playlist.title
        tracks = list(map(YMTrack.from_track, playlist.tracks))
        return YMPlaylist(uri=uri, name=name, tracks=tracks)


class YMTrack(Track):
    @staticmethod
    def from_track(track: YMTrack):
        uri = f"yandexmusic:track:{track.track_id}"
        name = track.title
        length = track.duration_ms
        artists = list(map(YMArtist.from_artist, track.artists))
        artwork = track.cover_uri
        return YMTrack(uri=uri, name=name, length=length, artwork=artwork, artists=artists)

    artwork = fields.String()


class YMArtist(Artist):
    @staticmethod
    def from_artist(artist: YArtist):
        uri = f"yandexmusic:artist:{artist.id}"
        name = artist.name,
        return YMArtist(uri=uri, name=name[0])
